;;;  -*- mode: LISP; Package: CL-USER; Syntax: COMMON-LISP;  Base: 10 -*-
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; 
;;; Author      : Dan Bothell 
;;; Address     : Carnegie Mellon University
;;;             : Psychology Department
;;;             : Pittsburgh,PA 15213-3890
;;;             : db30+@andrew.cmu.edu
;;; 
;;; Copyright   : (c)2002-2004 Dan Bothell
;;; Availability: Covered by the GNU LGPL, see LGPL.txt
;;; 
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; 
;;; Filename    : hook-fns.lisp
;;; Version     : 1.5
;;; 
;;; Description : NO system dependent code.
;;;             : This file contains the functions and variables 
;;;             : for managing the handlers connected to the ACT-R
;;;             : and R/PM hook functions as well as the redefinitions
;;;             : of the ACT-R and R/PM code to get my list style handlers.
;;; Bugs        : 
;;; 
;;; Todo        :
;;; 
;;; ----- History -----
;;;
;;; 05/10/2002  Dan
;;;             : Added this header
;;; 09/25/2002  Dan
;;;             : Updated with the newest ACT-R code (v5.0.6)
;;; 10/01/2002  Dan
;;;             : Updated version to 1.1 and fixed the packaging
;;;             : for building a standalone in ACL.
;;; 04/22/2004  Dan
;;;             : Updated the master-process class to version 2.2
;;;             : and added the LGPL license stuff.
;;; 4/22/2004   Dan [1.5]
;;;             : Added the license info.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 
;(in-package :cl-user)

;#+(and :allegro-ide (not :ACTR-ENV-ALONE)) (in-package :cg-user)

;;; lists of handlers to call for each of the hooks

(defvar *pre-hook-list* nil)
(defvar *post-hook-list* nil)
(defvar *enqueue-hook-list* nil)
(defvar *instan-hook-list* nil)
(defvar *reset-hook-list* nil)
(defvar *init-hook-list* nil)
(defvar *step-hook-list* nil)

;;; For the ACT-R I need to add the list hooks as globals

(defvar *conflict-set-hook-fns* nil
  "A list of functions that are called when *conflict-set-hook-fn* is called
   and having the same effect - the last instantiation returned by any of 
   these is the one used.")

(defvar *init-hook-fns* nil
  "A list of functions to be called when *init-hook-fn* gets called.")

(defvar *reset-hook-fns* nil
  "A list of functions to be called after executing a reset or reload.")

;;; Sometimes a function may be called by multiple hooks, and it'd be
;;; nice to know which one it was currently called by...

(defvar *which-hook-fn* nil)

;;; generic-hook-update-fn
;;; this function takes 2 parameters, hook-list is a list
;;; of handlers to update and arg is the argument passed to
;;; the hook function.  It calls the update-handler for each
;;; of the handlers on the list.  Most of the hooks can be
;;; handled by this function.

(defun generic-hook-update-fn (hook-list hook arg)
  (setf *which-hook-fn* hook)
  (dolist (x hook-list)
      (update-handler x arg)))

;;; instan-hook-update-fn
;;; This function is used to handle the conflict-set-hook-fns
;;; because they're special.  If any of the functions called
;;; on the list by ACT-R returns an instantiation, then that
;;; instantiation is the one that will fire.  So I have to check
;;; all the handlers to see if any return an instantiation, and 
;;; then return that if so.

(defun instan-hook-update-fn (instan)
  (setf *which-hook-fn* :instan)
  (let ((ret nil))
    (dolist (x *instan-hook-list* ret)
      (let ((val (update-handler x instan)))
        (when val
          (setf ret val))))))

;;; init-hook-update-plus-fn
;;; This is the function that gets called on the init-hook
;;; of ACT-R (when clear-all gets called).  It calls all of
;;; the handlers that are tied to the init-hook and then makes
;;; sure that the hooks that ACT-R clears get set again.

(defun init-hook-update-plus-fn ()
  (setf *which-hook-fn* :init)
  (dolist (x *init-hook-list*)
    (update-handler x nil))
  (push #'instan-hook-update-fn *conflict-set-hook-fns*)
  (push #'(lambda () 
          (generic-hook-update-fn *reset-hook-list* :reset nil))
      *reset-hook-fns*))  


;;; This is the redefinitions necessary to implement my hook lists
;;; First, from master-process.lisp add some new slots to the master-process
;;; class.

(defclass master-process ()
  ((schedule-queue :accessor sched-q  :initarg sched-q :initform nil)
   (clock :reader mp-time :writer clock :initarg :clock :initform 0.0)
   (is-slave :accessor slave-p :initarg :slave-p :initform nil)
   (trace-modules :accessor trace-modules :initarg :trace-mod :initform t)
   (randomize-time :accessor randomize-time :initarg :rand-time :initform nil)
   (param-lis :accessor param-lis :initarg :param-lst :initform nil)
   (auto-dequeue-p :accessor auto-dequeue-p :initarg :auto-dequeue-p
                   :initform nil)
   (saved-hook-fct :accessor saved-hook-fct :initarg :saved-hook-fct
                   :initform nil)
   (real-time-p :accessor real-time-p :initarg :real-time-p :initform nil)
   (output-queue :accessor output-q :initarg output-q :initform nil)  
   (sim-start-real :accessor sim-start-real :initarg :sim-start-real 
                   :initform 0)
   (start-time :accessor start-time :initform 0.0)
   (version-string :accessor version-string :initarg :version-string
                   :initform "2.2")
   (module-lst :accessor module-lst :initarg :module-lst :initform nil)
   (enqueue-hook :accessor enqueue-hook :initarg :enqueue-hook :initform nil)
   (pre-event-hook :accessor pre-event-hook :initarg :pre-event-hook 
                   :initform nil)
   (post-event-hook :accessor post-event-hook :initarg :post-event-hook 
                    :initform nil)
   (stepper-menu-p :accessor stepper-menu-p :initarg :stepper-menu-p 
                   :initform t)
   (step-hook :accessor step-hook :initarg :step-hook :initform nil)
   (rt-slack-hook-fct :accessor rt-slack-hook-fct :initarg :rt-slack-hook-fct
                      :initform nil)
   
   ;;; Additions for the environment support
   (enqueue-hooks :accessor enqueue-hooks :initarg :enqueue-hooks 
                  :initform nil)
   (pre-event-hooks :accessor pre-event-hooks :initarg :pre-event-hooks 
                   :initform nil)
   (post-event-hooks :accessor post-event-hooks :initarg :post-event-hooks 
                    :initform nil)
))


;;; and change the scheduler so that it calls the new list hooks

(defmethod run-sched-queue ((mstr-proc master-process) stop-time)
  (when *debug* (check-queue-integ (sched-q mstr-proc)))
  (while (and (sched-q mstr-proc)
              (>= (ms-round stop-time)
                  (ms-round (time-tag (first (sched-q mstr-proc))))))
    (let ((next-entry (pop (sched-q mstr-proc))))
      (set-clock mstr-proc (time-tag next-entry))
      (when (functionp (pre-event-hook mstr-proc))
        (funcall (pre-event-hook mstr-proc) next-entry))
      (dolist (x (pre-event-hooks mstr-proc))
        (when (functionp x)
          (funcall x next-entry)))
      (case (destination next-entry)
        (:EXTERNAL 
         (apply (first (params next-entry)) (rest (params next-entry))))
        (:DEVICE
         (run-events (device-interface mstr-proc) (mp-time mstr-proc)))
        (otherwise
         (run-module (key->mgr mstr-proc (destination next-entry))
                     (mp-time mstr-proc))))
      (when (functionp (post-event-hook mstr-proc))
        (funcall (post-event-hook mstr-proc) next-entry))
      (dolist (x (post-event-hooks mstr-proc))
        (when (functionp x)
          (funcall x next-entry))))))

(defmethod new-message ((mstr-proc master-process) (entry queue-entry))
  (setf (time-tag entry) (+ (mp-time mstr-proc) (time-tag entry)))
  (when (not (member entry (sched-q mstr-proc) :test #'sched=))
    (when (functionp (enqueue-hook mstr-proc))
      (funcall (enqueue-hook mstr-proc) entry))
    (dolist (x (enqueue-hooks mstr-proc))
        (when (functionp x)
          (funcall x entry)))
    (setf (sched-q mstr-proc)
          (queue-insert (make-instance 'queue-entry :time-tag (time-tag entry)
                          :priority (priority entry)
                          :destination (destination entry))
                        (sched-q mstr-proc))))
  (new-message (key->mgr mstr-proc (destination entry)) entry))

;;; That's it for the RPM redefinintions, and here are the ACT-R ones

;;; Then modify clear-all so it properly clears them and calls the 
;;; init-hooks

(defun clear-all-fct (&optional (save-model t))
  "Clears everything."
  (clear-used-symbols-fct)
  (setf *compile-eval-calls* nil)
  (when *load-pathname*
    (setf *model* (when save-model (list *load-pathname*)))
    (when (eq save-model t)
      (load-model-list *load-pathname*)))
  (setf *time* 0.0)
  (setf *start-time* 0.0)
  (setf *default-action-time* 0.05)
  (setf *latency* 0.0)
  (setf *cycle* 0)
  (setf *spread-stamp* 0)
  (setf *wme-number* 0.0)
  (setf *buffers* nil)
  (setf *declarative-memory* nil)
  (setf *hash-names* (make-hash-table :test #'eq :size 1000))
  (init-types)
  (setf *procedural-memory* nil)
  (setf *goal-activation* 1.0)
  (setf *goal-sources* nil)
  (setf *wmfocus* nil)
  (setf *wmfocus-list* nil)
  (setf *goal-stack* nil)
  (setf *goal-depth* 1)
  (setf *g* 20.0)
  (setf *exp-gain-noise* nil)
  (setf *retrieval* nil)
  (setf *retrieval-scheduler* nil)
  (setf *visual-location* nil)
  (setf *aural-location* nil)
  (setf *previous-instantiations* nil)
  (setf *previous-instantiation* nil)
  (setf *instantiation* nil)
  (setf *extra-instantiation* nil)
  (setf *conflict-set* nil)
  (setf *sum-exp-act* 0.0)
  (setf *latency-fn* 'old-latency)
  (setf *latency-factor* 1.0)
  (setf *latency-exponent* 1.0)
  (setf *base-level-constant* 0.0)
  (setf *activation-sources* nil)
  (setf *activation-noise* nil)
  (setf *permanent-activation-noise* nil)
  (setf *mismatch-penalty* 1.0)
  (setf *retrieval-threshold* nil)
  (setf *partial-matching* nil)
  (setf *max-sim* 0.0)
  (setf *max-dif* -1.0)
  (setf *maximum-associative-strength* nil)
  (setf *temperature* 1.0)
  (setf *blending* nil)
  (setf *break-productions* nil)
  (setf *failed-productions* nil)
  (setf *cost-penalty* 1.0)
  (setf *initial-experience* 10.0)
  (setf *threshold-time* 2.0)
  (setf *enable-production-learning* nil)
  (setf *reinforce-analogized-production* t)
  (setf *production-compilation-parameters* nil)
  (setf *enable-rational-analysis* nil)
  (setf *enable-randomness* nil)
  (setf *utility-threshold* 0.0)
  (setf *optimized-learning* t)
  (setf *base-level-learning* nil)
  (setf *associative-learning* nil)
  (setf *strength-learning* nil)
  (setf *parameters-learning* nil)
  (setf *command-trace* t)
  (setf *output-trace* t)
  (setf *cycle-trace* t)
  (setf *latency-trace* nil)
  (setf *partial-matching-trace* nil)
  (setf *blending-trace* nil)
  (setf *production-compilation-trace* t)
  (setf *activation-trace* nil)
  (setf *conflict-resolution-trace* nil)
  (setf *conflict-set-trace* nil)
  (setf *goal-trace* nil)
  (setf *dm-trace* nil)
  (setf *production-trace* nil)
  (setf *matches-trace* nil)
  (setf *exact-matching-trace* nil)
  (setf *verbose* t)
  (setf *abort-instantiation* t)
  (setf *parameter-sets* nil)
  (setf *similarity-hook-fn* nil)
  (setf *blending-hook-fn* 'blending-arithmetic-mean)
  (setf *conflict-set-hook-fn* nil)
  (setf *conflict-set-hook-fns* nil)
  (setf *firing-hook-fn* nil)
  (setf *cycle-hook-fn* nil)
  (setf *web-hook-fn* nil)
  (setf *end-run-hook-fn* nil)
  (setf *reset-hook-fns* nil)
  (setf *step-fn* 'step-fct)
  (setf *stop* nil)
  (setf *save-state-changes* nil)
  (setf *pop-upon-failure* nil)
  (when *init-hook-fn* (funcall *init-hook-fn*))
  (when *init-hook-fns* 
    (dolist (x *init-hook-fns*)
      (funcall x)))
  )

;;; add the reset hook to the reset and reload functions

(defun reset-fct (&optional (seed t))
  "Resets the current model by reevaluating all the commands."
  (cond ((rest *model*)
         (seed seed)
         (dolist (command (rest *model*))
           (eval command)))
        (t
         (reload-fct seed)))
  (when *reset-hook-fns*
    (dolist (x *reset-hook-fns*)
      (funcall x))))

(defun reload-fct (&optional (seed t))
  "Updates the current model by reloading the file."
  (cond 
   ((first *model*)
    (seed seed)
    (load (first *model*) :verbose *verbose*))
   (t
    (signal-warn "NO MODEL STORED.  LOAD THE MODEL FILE BEFORE USING RESET ~
                  OR RELOAD.~%DO NOT LOAD MODEL BY EVALUATING THE BUFFER.")))
  (when *reset-hook-fns*
    (dolist (x *reset-hook-fns*)
      (funcall x))))

;;; add the conflict-set hooks to the matching-cycle

(defun matching-cycle (&optional (cycles nil) 
                                 (retrieval-threshold *retrieval-threshold*) 
                                 (utility-threshold *utility-threshold*))
  "Implements the production matching (lhs) cycle called by run-fct and the 
   PM scheduler.  Cycles is the time limit for the matching cycle if 
   applicable. Returns the latency of the action phase if any, nil otherwise."
  (let ((matching-latency 0.0)
        (failed-latencies 0.0)
        (action nil))
    (when (and *retrieval-threshold* *activation-noise* *threshold-noise*)
      (let ((noise (noise *activation-noise*)))
        (setf *retrieval-threshold* (+ retrieval-threshold noise))
        (signal-output *activation-trace* 
                       "   Adding noise ~6,3F to retrieval threshold for a ~
                           total of ~6,3F"
                       noise *retrieval-threshold*)))
    (when (and *utility-threshold* *exp-gain-noise* *threshold-noise*)
      (let ((noise (noise *exp-gain-noise*)))
        (setf *utility-threshold* (+ utility-threshold noise))
        (signal-output (or *exact-matching-trace* *partial-matching-trace*)
                       "   Adding noise ~6,3F to utility threshold for a ~
                        total of ~6,3F"
                       noise *utility-threshold*)))
    (generate-all-instantiations)
    (when *matches-trace*
      (pprint-instantiations *conflict-set* *matches-trace* *matches-trace*))
    (setf *instantiation* nil)
    (when (or *conflict-set-hook-fn* *conflict-set-hook-fns*)
      (let ((instantiations (if *conflict-set-hook-fn*
                                (funcall *conflict-set-hook-fn* 
                                         *conflict-set*)
                              nil)))
        (when *conflict-set-hook-fns*
          (dolist (x *conflict-set-hook-fns*)
            (let ((val (funcall x *conflict-set*)))
              (when val (setf instantiations val)))))
        
        (when instantiations ; if returns nil, then continue unchanged
          (cond ((listp instantiations) ; if returns a list, 
                                        ; then interpret as new cset
                 (setf *conflict-set* instantiations))
                (t ; otherwise, restrict the conflict set to 
                   ; selected instantiation
                 (setf *conflict-set* (list instantiations)))))))
    (unless *instantiation*        
      (multiple-value-setq (*instantiation* matching-latency failed-latencies)
        (choose-instantiation)))
    ;;; the following used to be part of the action cycle but really 
    ;;; belongs in matching
    (cond (*instantiation*
           (let ((production (instantiation-production *instantiation*)))
             (cond ((and *abort-instantiation* *enable-rational-analysis*
                         (floatp cycles)
                         (> (+ *time* matching-latency)
                            cycles))
                    (signal-output *latency-trace* 
                                   "Time ~6,3F: ~A Aborted" 
                                   *time* production))
                   (t
                    (when (or *base-level-learning* *associative-learning*
                              *strength-learning*)
                      (learn-matching))
                    (when *enable-rational-analysis*
                      (incf *time* matching-latency)
                      (when (plusp matching-latency)
                        (signal-output *latency-trace* 
                                       "Latency ~6,3F: ~A Matching"
                                       matching-latency production)))
                    (signal-output *cycle-trace* 
                                   "Time ~6,3F: ~A Selected" 
                                   *time* production)
                    (setf (instantiation-latency *instantiation*)
                          (- matching-latency failed-latencies))
                    (setf action (production-action-cost production))))))
          ((and *enable-rational-analysis* 
                *pop-upon-failure* (null *retrieval-scheduler*))
           (incf *time* matching-latency)
           (when (plusp matching-latency)
             (signal-output *latency-trace* "Latency ~6,3F: Failure Matching"
                            matching-latency))
           (setf action *default-action-time*))
          (t
           (signal-output *latency-trace* 
                          "Time ~6,3F: No Instantiation Found." *time*)))
    action))


;;; end of redefinitions


;;; Here I put all of the handler hook functions in place for the environment

;;; The RPM hooks just need to be set up once as does the init hook

(push #'(lambda (x) 
          (generic-hook-update-fn *pre-hook-list* :pre x)) 
      (pre-event-hooks *mp*))
(push #'(lambda (x) 
          (generic-hook-update-fn *post-hook-list* :post x))
      (post-event-hooks *mp*))
(push #'(lambda (x) 
          (generic-hook-update-fn *enqueue-hook-list* :enqueue x))
      (enqueue-hooks *mp*))

(push #'init-hook-update-plus-fn *init-hook-fns*)

(setf (step-hook *mp*) #'(lambda (x)
                           (generic-hook-update-fn *step-hook-list* :step x)))


;;; The conflict set and reset hooks get set up by the init hook
;;; but doing it here just for completeness

(push #'instan-hook-update-fn *conflict-set-hook-fns*)

(push #'(lambda () 
          (generic-hook-update-fn *reset-hook-list* :reset nil))
      *reset-hook-fns*)

#|
This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
|#